/*
 * fepg.c: An EPG plugin for the Video Disk Recorder
 *
 * See the README file for copyright information and how to reach the author.
 *
 */

#include <vdr/plugin.h>
#include <vdr/interface.h>
#include <string>
#include <vector>
#include <sstream>
#include <stdlib.h>
#include <time.h>

#ifndef VDRVERSNUM
#include <vdr/config.h>
#endif

#include "fepg.h"
#include "menu.h"
#include "config.h"
#include "tools.h"


using namespace std;



static const char *VERSION        = "0.1.0";
static const char *DESCRIPTION    = "A Graphical EPG";
static const char *MAINMENUENTRY  = "fEPG";


//////////////////////////////////////////////////////////////////////////////


cFEpg::cFEpg(void)
{
     osd = NULL;
     width = SCR_WIDTH - 2*DX;
     height = SCR_HEIGHT - DY - DYO;   
     cWidth = (width - NAME_W)/3;
     info_open = false;
     firstTime = new struct tm;
     curEvent = NULL;
     curCellh = 1;
     curCellv = 1;
     inputTime = 0;
     
#if VDRVERSNUM>=10300
     font = cFont::GetFont(fontOsd);
#endif
}

//----------------------------------------------------------------------------

cFEpg::~cFEpg()
{
  if (osd != NULL)
    delete osd;
  if (bitmap != NULL)
    delete bitmap;
  if (firstTime != NULL)
    delete firstTime;
}

//----------------------------------------------------------------------------

void cFEpg::draw(int start_sid, char direction = 'D')
{
     time_t times[3];

     // Clear the event matrix
     for (int i=0; i<NUM; i++)
       for (int j=0; j<3; j++)
       {
         eventMtx[i][j] = NULL;
       }

     vector<const cSchedule*> ScheduleList;
     
#if VDRVERSNUM>=10300
     cSchedulesLock SchedulesLock;
     const cSchedules* Schedules = cSchedules::Schedules(SchedulesLock);
#else 
     cMutexLock mutexLock;
     const cSchedules* Schedules = cSIProcessor::Schedules(mutexLock);
#endif

     // Check if channel is valid
     bool ok = false;
     while(!ok)
     {
       if (start_sid < 0) direction = 'D';
       if (start_sid > Channels.MaxNumber()) direction = 'U';
       cChannel* check = Channels.GetByNumber(start_sid);
       if (check)
         ok=true;
       else
       {
         if (direction == 'U')
           start_sid--;
         else
           start_sid++;
       }    
     }  
     
     currentChannel = start_sid;

     // Get current time
     time_t rawtime;
     time(&rawtime);
     currentTime = localtime(&rawtime);

     // Update firstTime if necessary - DOESN'T WORK
     if ((mktime(firstTime) - mktime(currentTime))/60 < -30)
     {
       // Add 0:30 to firstTime
       if (firstTime->tm_min == 30)
       { 
         firstTime->tm_min = 0;
         firstTime->tm_hour += 1;
       }
       else
         firstTime->tm_min = 30;
     }


     
     
#if VDRVERSNUM>=10300
     bitmap->DrawRectangle(0,0,720,480,clrBlack);   
     bitmap->DrawRectangle(DX+NAME_W, DYO, DX+width, DYO+height, bgColor);
     bitmap->DrawRectangle(DX, DYO, DX+NAME_W, DYO+height, green);
     bitmap->DrawRectangle(DX, DY, DX+width, DYO, green);
#else
     bitmap->Fill(0,0,720,480,clrBlack);   
     bitmap->Fill(DX+NAME_W, DYO, DX+width, DYO+height, bgColor);
     bitmap->Fill(DX, DYO, DX+NAME_W, DYO+height, green);
     bitmap->Fill(DX, DY, DX+width, DYO, green);
#endif
     
     // Draw current time
     const char* s = toTimeString(currentTime).c_str();
     Text(DX+(NAME_W-(Width(s)))/2, DY+10, s, lightgrey, green);



     // Draw the other times
     times[0] = mktime(firstTime);
     s = toTimeString(firstTime).c_str();
     Text(DX+NAME_W+(cWidth-Width(s))/2, DY+10, s, lightgrey, green);


     struct tm* t = new struct tm;
     (*t) = (*firstTime);

     if (firstTime->tm_min == 30)
     {
       t->tm_hour = firstTime->tm_hour + 1;
       t->tm_min = 0;
       s = toTimeString(t).c_str();
       
       Text(DX+NAME_W+cWidth+(cWidth-Width(s))/2, DY+10, s, lightgrey, green);


       times[1] = mktime(t);

       t->tm_min = 30;
       s = toTimeString(t).c_str();
       Text(DX+NAME_W+2*cWidth+(cWidth-Width(s))/2, DY+10, s, lightgrey, green);

       times[2] = mktime(t); 
     }
     else
     {
       t->tm_hour = firstTime->tm_hour;
       t->tm_min = 30;
       s = toTimeString(t).c_str();
       Text(DX+NAME_W+cWidth+(cWidth-Width(s))/2, DY+10, s, lightgrey, green);
       times[1] = mktime(t);

       t->tm_hour = firstTime->tm_hour + 1;
       t->tm_min = 0;
       s = toTimeString(t).c_str();
       Text(DX+NAME_W+2*cWidth+(cWidth-Width(s))/2, DY+10, s, lightgrey, green);
       times[2] = mktime(t);       
     }
     delete t;

     // Write the channel numbers & names
     int numberOfChans = 0;
     int n = 0;
     while(numberOfChans < NUM)
     {
        // Make sure channels don't get too big...
        if (start_sid + n > Channels.MaxNumber())
        {
          draw(--currentChannel, 'U');
          return;
        }

        const char* chanNum = (to_string(start_sid + n)).c_str();
        cChannel* c = Channels.GetByNumber(start_sid + n);
        
        // Make sure this is a valid channel
        if (c)
        {
          displayedChannels[numberOfChans] = start_sid + n;
          const char* chanName = c->Name();
          Text(DX+(NAME_W-(Width(chanNum)))/2, DYO+5+numberOfChans*(height/NUM), chanNum, lightgrey, green);
          Text(DX+(NAME_W-(Width(chanName)))/2, DYO+30+numberOfChans*(height/NUM), chanName, lightgrey, green);
          numberOfChans++;

          // Get Schedule for this channel
          ScheduleList.push_back(Schedules->GetSchedule(c->GetChannelID()));
        }
           
        n++;
     }

     
     curEvent = NULL;
    

     //--- Put events in EPG -----
     for (int j=0; j<NUM; j++)
     {
       time_t stTime = times[0];
       long delTime = 3*30*60;
       while (stTime <= times[2])
       {
         int barX = width - NAME_W;  // place for the seperator: default value
#if VDRVERSNUM>=10300
         const cEvent* event = NULL;
#else         
         const cEventInfo* event = NULL;
#endif         

         const char* title = "No Info";
         if (ScheduleList.at(j))
         {
           event = ScheduleList.at(j)->GetEventAround(stTime);
           if (event)
           {
#if VDRVERSNUM>=10300
             title = event->Title();
             delTime = event->StartTime() + event->Duration() - stTime;
#else           
             title = event->GetTitle();
             delTime = event->GetTime() + event->GetDuration() - stTime;
#endif             
             
             // Calculate place for the seperator
             barX = static_cast<int>((static_cast<float>(delTime/60))/30.0 * cWidth);
           }
         }
       
         int cellBuf = 0; // How far away (in X direction) to start drawing event


         if (stTime == times[0])
         {
           if (barX <= cWidth)
           {             
             barX = cWidth;
             eventMtx[j][0] = event;
           }
           else if (barX <= 2*cWidth)
           {
             barX = 2*cWidth;
             eventMtx[j][0] = event;
             eventMtx[j][1] = event;
           }
           else 
           {            
             barX = width - NAME_W;
             eventMtx[j][0] = event;
             eventMtx[j][1] = event;
             eventMtx[j][2] = event;
           }
         }
         else if (stTime == times[1])
         {
           cellBuf = cWidth;
           if (barX <= cWidth)
           {
             barX = cWidth; 
             eventMtx[j][1] = event;
           }
           else
           {
             barX = 2*cWidth;
             eventMtx[j][1] = event;
             eventMtx[j][2] = event;
           }
         }
         else if (stTime == times[2])
         {
           cellBuf = 2*cWidth;
           barX = cWidth;
           eventMtx[j][2] = event;
         }
         
#if VDRVERSNUM>=10300
         tColor clr1 = lightgrey;
         tColor clr2 = bgColor;
#else
         eDvbColor clr1 = lightgrey;
         eDvbColor clr2 = bgColor;
#endif         
       
         bool selected = false;
         // Check if this cell is selected
         if (eventMtx[j][curCellv-1] == event && j == curCellh-1) 
         {    
           selected = true;
           curEvent = event;
           clr1 = bgColor;
           clr2 = lightgrey;
           
           //Fill 
#if VDRVERSNUM>=10300
          bitmap->DrawRectangle(cellBuf+DX+NAME_W, DYO+j*(height/NUM), cellBuf+DX+NAME_W+barX, DYO+(j+1)*(height/NUM), lightgrey);
#else
           bitmap->Fill(cellBuf+DX+NAME_W, DYO+j*(height/NUM), cellBuf+DX+NAME_W+barX, DYO+(j+1)*(height/NUM), lightgrey); 
#endif    
         }  

         int availableWidth = barX;
         int offSet = 7;

         // Draw Arrow if event has already started or will not end
         if (event != NULL)
         {
#if VDRVERSNUM>=10300
           if (event->StartTime()+30*60 < times[0]) 
           {
             Text(cellBuf+DX+NAME_W+6, DYO+20+j*(height/NUM), "<", clr1, clr2);
             offSet = Width("<") - 3;
             availableWidth -= offSet;
           }
           if (event->StartTime() + event->Duration() > (times[2]+60*30))
           {
             Text(DX+width-15, DYO+20+j*(height/NUM), ">", clr1, clr2);
             availableWidth -= Width(">");
             offSet = 0;
           }
#else
           if (event->GetTime()+30*60 < times[0]) 
           {
             bitmap->Text(cellBuf+DX+NAME_W+6, DYO+20+j*(height/NUM), "<", clr1, clr2);
             offSet = osd->Width("<") - 3;
             availableWidth -= offSet;
           }
           if (event->GetTime() + event->GetDuration() > (times[2]+60*30))
           {
             bitmap->Text(DX+width-15, DYO+20+j*(height/NUM), ">", clr1, clr2);
             availableWidth -= osd->Width(">");
             offSet = 0;
           }
#endif    
         }    

         vector<string> wrappedTitle;
         string t(title);
         if (event != NULL)
         {
	         // If event doesn't start at :00 or :30, add start time as prefix
		       string st(event->GetTimeString());
		       string st2 = st.substr(st.find_first_of(":")+1);
	         if (st2 != "00" && st2 != "30")
           {
             st = to12time(st);
	           t = "("+ st +") "+ t;
           }
         }
         WrapText(t, wrappedTitle, availableWidth - 15, false);
        

         // Draw name of event
         if (wrappedTitle.size() == 1)
         {
           Text(cellBuf+DX+NAME_W+offSet+(barX-Width(wrappedTitle.at(0).c_str()))/2, DYO+20+j*(height/NUM), wrappedTitle.at(0).c_str(), clr1, clr2);
         }
         else if (wrappedTitle.at(0) == "" && wrappedTitle.size() > 1)
         {
           if (wrappedTitle.size() == 2)
             Text(cellBuf+DX+NAME_W+offSet+(barX-Width(wrappedTitle.at(1).c_str()))/2, DYO+20+j*(height/NUM), wrappedTitle.at(0).c_str(), clr1, clr2);
           else
           {
             Text(cellBuf+DX+NAME_W+offSet+(barX-Width(wrappedTitle.at(1).c_str()))/2, DYO+5+j*(height/NUM), wrappedTitle.at(0).c_str(), clr1, clr2);
             Text(cellBuf+DX+NAME_W+offSet+(barX-Width(wrappedTitle.at(2).c_str()))/2, DYO+30+j*(height/NUM), wrappedTitle.at(1).c_str(), clr1, clr2);
           }
         }
         else //TO DO: expand to cover more cases
         {
           Text(cellBuf+DX+NAME_W+offSet+(barX-Width(wrappedTitle.at(0).c_str()))/2, DYO+5+j*(height/NUM), wrappedTitle.at(0).c_str(), clr1, clr2);
           Text(cellBuf+DX+NAME_W+offSet+(barX-Width(wrappedTitle.at(1).c_str()))/2, DYO+30+j*(height/NUM), wrappedTitle.at(1).c_str(), clr1, clr2);
         }

         if (selected) // Draw a box around selected event
         {      		       
           Fill(cellBuf+DX+NAME_W+3, DYO+j*(height/NUM), cellBuf+DX+NAME_W+4, DYO+(j+1)*(height/NUM), clrBlack);  //Left
           Fill(cellBuf+DX+NAME_W+barX-1, DYO+j*(height/NUM), cellBuf+DX+NAME_W+barX+1, DYO+(j+1)*(height/NUM), clrBlack);  //Right
           Fill(cellBuf+DX+NAME_W+3, DYO+j*(height/NUM)+3, cellBuf+DX+NAME_W+barX-1, DYO+j*(height/NUM)+4, clrBlack);   // Top
           Fill(cellBuf+DX+NAME_W+3, DYO+(j+1)*(height/NUM)-1, cellBuf+DX+NAME_W+barX-1, DYO+(j+1)*(height/NUM), clrBlack);  // Bottom 

         }
        
         // Draw seperator bar
#if VDRVERSNUM>=10300
         bitmap->DrawRectangle(cellBuf+DX+NAME_W+barX, DYO+j*(height/NUM), cellBuf+DX+NAME_W+barX+3, DYO+(j+1)*(height/NUM), lightgrey); 
#else         
         bitmap->Fill(cellBuf+DX+NAME_W+barX, DYO+j*(height/NUM), cellBuf+DX+NAME_W+barX+3, DYO+(j+1)*(height/NUM), lightgrey); 
#endif

         if (event)
         {
           stTime += delTime;

           if (stTime > times[0] && stTime <= times[1]) stTime = times[1];
           else if (stTime > times[1] && stTime <= times[2]) stTime = times[2];
           else if (stTime > times[2]) stTime = times[2] + 1000;
         }
         else // TO DO: What if there is no info now but there is some later
         {
           stTime = times[2] + 1000;
         }

       }
     }
      
     //---------------------------

     // Vertical lines
#if VDRVERSNUM>=10300
     bitmap->DrawRectangle(DX, DY, DX+3, DYO+(height/NUM)*NUM, lightgrey); //Vert Left
     bitmap->DrawRectangle(DX+width, DY, DX+width+3, DYO+(height/NUM)*NUM, lightgrey); //Vert right
     bitmap->DrawRectangle(DX+NAME_W, DY, DX+NAME_W+3, DYO+(height/NUM)*NUM, lightgrey);
#else     
     bitmap->Fill(DX, DY, DX+3, DYO+(height/NUM)*NUM, lightgrey); //Vert Left
     bitmap->Fill(DX+width, DY, DX+width+3, DYO+(height/NUM)*NUM, lightgrey); //Vert right
     bitmap->Fill(DX+NAME_W, DY, DX+NAME_W+3, DYO+(height/NUM)*NUM, lightgrey);
#endif     
     for (int i=1; i<3; i++)
     {
#if VDRVERSNUM>=10300
          bitmap->DrawRectangle(DX+NAME_W+i*cWidth, DY, DX+NAME_W+i*cWidth+3, DYO, lightgrey);
#else
          bitmap->Fill(DX+NAME_W+i*cWidth, DY, DX+NAME_W+i*cWidth+3, DYO, lightgrey);
#endif 
     }

#if VDRVERSNUM>=10300
     bitmap->DrawRectangle(DX, DY, DX+width+3, DY+3, lightgrey);  // Top
#else     
     bitmap->Fill(DX, DY, DX+width+3, DY+3, lightgrey);  // Top
#endif
     // Horizontal lines
     for (int k=0; k<=NUM; k++)
     {
#if VDRVERSNUM>=10300 
          bitmap->DrawRectangle(DX, DYO+(height/NUM)*k, DX+width+3, DYO+(height/NUM)*k+3, lightgrey);
#else   
          bitmap->Fill(DX, DYO+(height/NUM)*k, DX+width+3, DYO+(height/NUM)*k+3, lightgrey);
#endif          
     }

#if VDRVERSNUM>=10300 
     osd->DrawBitmap(0, 0, *bitmap);
#else
     osd->SetBitmap(0,0,*bitmap, windowHandle);
#endif
     osd->Flush();
}

//----------------------------------------------------------------------------

void cFEpg::info(const 
#if VDRVERSNUM>=10300
  cEvent
#else
  cEventInfo
#endif
* e)
{
  if (e == NULL)
  {
    return;
  }

  info_open = true;
  int W = (int)(0.8 * SCR_WIDTH);  // Width
  int H = (int)(0.7 * SCR_HEIGHT); // Height
  int B = 10; // Border
  
  int Y = (SCR_HEIGHT-H)/2; // Y Coord. of upper left corner
  int X = (SCR_WIDTH-W)/2;  // X Coord. of upper left corner
 
  // The Info Box 
  Fill(X, Y, X+W, Y+H, green);
  Fill(X, Y, X+W, Y+2, clrBlack);     //top
  Fill(X, Y+H-2, X+W, Y+H, clrBlack); //bottom
  Fill(X, Y, X+2, Y+H, clrBlack);     //left
  Fill(X+W-2, Y, X+W, Y+H, clrBlack); //right

  // The program name
  Fill(X+B,Y+B, X+W-B, Y+50, lightgrey);
#if VDRVERSNUM>=10300
  const char* name = e->Title() ? strdup(e->Title()) : NULL;
#else
  const char* name = e->GetTitle() ? strdup(e->GetTitle()) : NULL;
#endif  
  Text(X+(W-(Width(name)))/2, Y+15, name, clrBlack, lightgrey);

  // The description box
  int TOP = Y+2*B+50-10;
  Fill(X+B, TOP, X+W-B, Y+H-B, lightgrey);
  Fill(X+B, Y+H-B-60, X+W-B, Y+H-B-58, clrBlack);
  
  string start(e->GetTimeString());
  string end(e->GetEndTimeString());
  if (start.at(0) == '0')
    start.at(0) = ' ';
  if (end.at(0) == '0')
    end.at(0) = ' ';

  start = to12time(start);
  end   = to12time(end);

#if VDRVERSNUM>=10300
  long lDuration = e->Duration();
#else
  long lDuration = e->GetDuration();
#endif  

  start = "Start  : " + start;
  end = "End   : " + end;
  string duration = "Duration : " + toTimeString(lDuration);
  int num = displayedChannels[curCellh-1];
  string channel = "Channel : " + to_string(num); 

  Text(X+2*B, Y+H-B-55, start.c_str(), clrBlack, lightgrey);
  Text(X+2*B, Y+H-B-30, end.c_str(), clrBlack, lightgrey);
  int side = Width(duration.c_str());
  Text(X+W-2*B-side, Y+H-B-55, duration.c_str(), clrBlack, lightgrey);
  Text(X+W-2*B-side, Y+H-B-30, channel.c_str(), clrBlack, lightgrey);

  // The description text  
  vector<string> lines;
#if VDRVERSNUM>=10300
  string desc(e->Description());
#else
  string desc(e->GetExtendedDescription());
#endif
  WrapText(desc, lines, W-6*B, true);

  int buf;
  unsigned int off;
  const char* subTitle;
  if (BEV)
  {
    subTitle = lines.at(0).c_str();
    off = 1;
  }
  else
  {
#if VDRVERSNUM>=10300
    subTitle = e->ShortText();
#else
    subTitle = e->GetSubtitle();
#endif
    off = 0;
  }

  if (strcmp(subTitle, "") == 0)
    buf = BEV ? -25 : 0;
  else
  {
    buf = BEV ? 0 : 25;
    Text(X+(W-Width(subTitle))/2, TOP+10, subTitle, clrBlack, lightgrey);
  }

  // Check if text is contained in the box
  // TO DO: the number 6 is only ok for default size, should calculate this.
  if (lines.size() > 6)
    lines.at(5) = lines.at(5).substr(0, lines.at(5).length()-4) + "...";

  for (unsigned int n=off; n<lines.size() && n<6; n++)
  {
    Text(X+3*B, TOP+10+buf+n*27, lines.at(n).c_str(), clrBlack, lightgrey); 
  }

#if VDRVERSNUM>=10300 
     osd->DrawBitmap(0, 0, *bitmap);
#else
     osd->SetBitmap(0,0,*bitmap, windowHandle);
#endif
     osd->Flush();

}

//----------------------------------------------------------------------------


void cFEpg::Show(void)
{
  // Mute
  if (!cDevice::PrimaryDevice()->IsMute())
  {
    wasMuted = false;
    cDevice::PrimaryDevice()->ToggleMute();
  }
  else
    wasMuted = true;
    
#if VDRVERSNUM>=10300
  osd = cOsdProvider::NewOsd(0, 0);
#else    
  osd = cOsd::OpenRaw(0, 0);
#endif

  if (osd) 
  {
 #if VDRVERSNUM>=10300
     tArea Area = { 0, 0, SCR_WIDTH, SCR_HEIGHT,  2 };
     
     if (osd->CanHandleAreas(&Area, 1) != oeOk)
     {
      fprintf(stderr, "fEPG: OSD Error #%d\n", osd->CanHandleAreas(&Area, 1));
     }
     osd->SetAreas(&Area, 1);

 #else
     windowHandle = osd->Create(0, 0, SCR_WIDTH, SCR_HEIGHT, 2, false);
     osd->AddColor(clrBlack);
     osd->AddColor(bgColor);
     osd->AddColor(green);
     osd->AddColor(lightgrey);    
 #endif

     // Get current time
     time_t rawtime;
     time(&rawtime);
     currentTime = localtime(&rawtime);
     
     (*firstTime) = (*currentTime);
     //Round down firstTime to nearest 30 min
     if ((*firstTime).tm_min > 30)     
       (*firstTime).tm_min = 30;
     else
       (*firstTime).tm_min = 0;

#if VDRVERSNUM>=10300 
     bitmap = new cBitmap(720,480,2);  
#else    
     bitmap = new cBitmap(720,480,2,false);
#endif
    
     currentChannel = cDevice::CurrentChannel();   
     draw(currentChannel);

  }
  else
  {
    fprintf(stderr, "fEPG: Failed to create OSD");
  }
  
}

//----------------------------------------------------------------------------

eOSState cFEpg::ProcessKey(eKeys Key)
{
  eOSState state = cOsdObject::ProcessKey(Key);
  if (state == osUnknown) 
  {
     
       // Move up by one channel
       if (NORMALKEY(Key) == config.keyUp)
       {
         if (!info_open)
         {
           if (curCellh == 1)
             draw(--currentChannel, 'U');
           else
           {
             curCellh--;
             draw(currentChannel);
           }
         }
       }
         
       // Move down by one channel
       else if (NORMALKEY(Key) == config.keyDown)
       {
         if (!info_open)
         {
           if (curCellh == NUM)
             draw(++currentChannel);
           else
           {
             curCellh++;
             draw(currentChannel);
           }
         }
       }

       // Scroll left by 0:30
       else if (NORMALKEY(Key) == config.keyLeft)
       {   
         if (!info_open)
         {
           //Make sure firstTime doesn't get too small
           if (!((mktime(firstTime) - mktime(currentTime))/60 <= 0 && curCellv == 1))
           {  

            if (curCellv == 1)
            {
              if (firstTime->tm_min == 30)
                firstTime->tm_min = 0;
              else
              {  
               firstTime->tm_hour -= 1;
               firstTime->tm_min = 30;
              } 
            }
            else
            {
              bool stop = false;
              
#if VDRVERSNUM>=10300 
              const cEvent* e = eventMtx[curCellh-1][curCellv-1];
#else
              const cEventInfo* e = eventMtx[curCellh-1][curCellv-1];
#endif              
              
              do
              { 
                if (curCellv == 1)
                {
                  ProcessKey((eKeys) config.keyLeft);
                  curCellv = 2;
                  stop = true;
                }
                curCellv--;
              } while(e == eventMtx[curCellh-1][curCellv-1] && !stop);
            }

             draw(currentChannel);
          }
         }
       }

       // Scroll right by 0:30
       else if (NORMALKEY(Key) == config.keyRight)
       {
         if (!info_open)
         {   
           if (curCellv == 3)
           {
             if (firstTime->tm_min == 30)
             { 
               firstTime->tm_min = 0;
               firstTime->tm_hour +=1;
             }
             else
               firstTime->tm_min = 30;
           }
           else
           {
              bool stop = false;
              
#if VDRVERSNUM>=10300
              const cEvent* e = eventMtx[curCellh-1][curCellv-1];
#else 
              const cEventInfo* e = eventMtx[curCellh-1][curCellv-1];
#endif  
            
              do
              {
                if (curCellv == 3)
                {
                  ProcessKey((eKeys)config.keyRight);
                  curCellv = 2;
                  stop = true;
                }
                curCellv++;
              } while(e == eventMtx[curCellh-1][curCellv-1] && !stop);
              
           }
             draw(currentChannel);
         }
       }

       // Open/close info window
       else if (Key == config.keyInfo)
       {
         if (info_open)
         {
           info_open = false;
           draw(currentChannel);
         }
         else
         {
           info(curEvent);  
         }   
       }  

       // Page up
       else if (NORMALKEY(Key) == config.keyPageUp)
       {
         if (!info_open)
         {
           draw(jumpTo(currentChannel, NUM, 'U'));
         }
       }

       // Page down
       else if (NORMALKEY(Key) == config.keyPageDown)
       {
         if (!info_open)
         {
           draw(jumpTo(currentChannel, NUM, 'D'));
         }
       }

       // Take screenshot
       else if (Key == config.keyScreenShot)
       {
         cDevice::PrimaryDevice()->GrabImage("img.jpg", true, 256, -1, -1);
       }

  /*     // Switch to this channel, gets EPG data much faster
       else if (Key == kUser2)
       {
         if (BEV)
         {
           Channels.SwitchTo(431);
         }
       }
*/
       // Exit EPG, no channel switch
       else if (Key == config.keyBack)
       {
         if (!info_open)
         {
           if (!wasMuted)
             cDevice::PrimaryDevice()->ToggleMute();
           return osEnd;
         }
         else
         {
           info_open = false;
           draw(currentChannel);
         }
       }
      
       //Go to channel#
       //TO DO: Rewrite this section, for more general situations
       else if (Key >= k0 && Key <= k9)
       {
         if (!info_open)
         {
           if (inputTime == 0)  inputTime = time_ms();
           if (time_ms() - inputTime > 1000)  inputNum.clear();
 
           inputNum.push_back(Key - k0);
           inputTime = time_ms();
         
           // Does the number have 3 digits?
           if (inputNum.size() == 3)
           {
             curCellh = 1;
             draw(100*inputNum.at(0)+10*inputNum.at(1)+inputNum.at(2));
             inputNum.clear();
           }
         }
       }

       //Switch channel and exit or close info window
       else if (Key == config.keyOk)
       {
         if (!info_open)
         {      
           // If the first event is selected, then switch
#if VDRVERSNUM>=10300    
           const cEvent* ev = eventMtx[curCellh-1][0];
           if (eventMtx[curCellh-1][curCellv-1] == ev
               && ev->StartTime() < mktime(currentTime)) 
#else     
           const cEventInfo* ev = eventMtx[curCellh-1][0];
           if (eventMtx[curCellh-1][curCellv-1] == ev
               && ev->GetTime() < mktime(currentTime))
#endif          

           {
             if (cDevice::CurrentChannel() != displayedChannels[curCellh-1])
             {
               Fill(0,0,SCR_WIDTH,SCR_HEIGHT,clrBlack);
               
#if VDRVERSNUM>=10300 
               osd->DrawBitmap(0, 0, *bitmap);
#else
               osd->SetBitmap(0,0,*bitmap, windowHandle);
#endif

               osd->Flush();
               Channels.SwitchTo(displayedChannels[curCellh-1]);
             }
             if (!wasMuted)
               cDevice::PrimaryDevice()->ToggleMute();
             return osEnd;
           }
           else  // Open info window
           {
             info(curEvent);  
           }
         }
         else
         {
           info_open = false;
           draw(currentChannel);
         }
       }
      
       else 
        return state;
       

     
     state = osContinue;
     }
  return state;
}


//////////////////////////////////////////////////////////////////////////////


void cFEpg::WrapText(string s, vector<string>& lines, int width, bool title)
{
  // Tokenize
  vector<string> tokens;

  if (s == "")
  {
    lines.push_back("");
    return;
  }

  string word = "";
  for (unsigned int i = 0; i < s.length(); i++)
  {
    if (s.at(i) != ' ')
      word += s.at(i);
    else
    {
      tokens.push_back(word);
      word = "";
    }
  }

  tokens.push_back(word);


  int total = 0;
  string temp = "";
  bool foundTitle = false;
  for (unsigned int k = 0; k < tokens.size(); k++)
  {
    if (BEV && title && !foundTitle)
    {
      string st = tokens.at(k);
      if (st=="Series." || st=="Movie." || st=="News." || st=="Education." || st=="Sports.")
      {
        lines.push_back(temp);
        temp = "";
        total = 0;
        
        foundTitle = true;
      }
    }
    if (Width(tokens.at(k).c_str()) + total >= width && temp != "")
    {
      lines.push_back(temp);
      temp = "";
      total = 0;
    }

    // Make sure the token is not too wide 
    while (Width(tokens.at(k).c_str()) > width)
    {
      string split = "";
      unsigned int idx = 0;
      while (Width(split.c_str()) <= width && idx < tokens.at(k).length())
      {
        split += tokens.at(k).at(idx++);
      }
      lines.push_back(split.substr(0, split.length()-2) + "-");
      tokens.at(k) = tokens.at(k).substr(idx-2);
    }

    temp += (tokens.at(k) + " ");
    total = Width(temp.c_str());
  }
  
  if (temp != "")
    lines.push_back(temp);

}

//----------------------------------------------------------------------------

#if VDRVERSNUM>=10300 
 
  void cFEpg::Text(int x, int y, const char* s, tColor c1, tColor c2)
  {
     bitmap->DrawText(x, y, s, c1, c2, font);
  }
  
  int cFEpg::Width(const char* s)
  {
    return font->Width(s);
  }
  
  void cFEpg::Fill(int x1, int y1, int x2, int y2, tColor c)
  {
    bitmap->DrawRectangle(x1, y1, x2, y2, c);     
  }
  
#else

  void cFEpg::Text(int x, int y, const char* s, eDvbColor c1, eDvbColor c2)
  {
    bitmap->SetFont(fontOsd);
    bitmap->DrawText(x, y, s, c1, c2);
  }
  
  int cFEpg::Width(const char* s)
  {
    osd->SetFont(fontOsd);
    return osd->Width(s);
  }
  
  void cFEpg::Fill(int x1, int y1, int x2, int y2, eDvbColor c)
  {
    bitmap->Fill(x1, y1, x2, y2, c);
  }
  
#endif


/// cPluginFepg //////////////////////////////////////////////////////////////

class cPluginFepg : public cPlugin {
private:
  // Add any member variables or functions you may need here.
public:
  cPluginFepg(void);
  virtual ~cPluginFepg();
  virtual const char *Version(void) { return VERSION; }
  virtual const char *Description(void) { return DESCRIPTION; }
  virtual const char *CommandLineHelp(void);
  virtual bool ProcessArgs(int argc, char *argv[]);
  virtual bool Start(void);
  virtual void Housekeeping(void);
  virtual const char *MainMenuEntry(void) { return MAINMENUENTRY; }
  virtual cOsdObject *MainMenuAction(void);
  virtual cMenuSetupPage *SetupMenu(void);
  virtual bool SetupParse(const char *Name, const char *Value);
  };
  
  


cPluginFepg::cPluginFepg(void)
{
  // Initialize any member variables here.
  // DON'T DO ANYTHING ELSE THAT MAY HAVE SIDE EFFECTS, REQUIRE GLOBAL
  // VDR OBJECTS TO EXIST OR PRODUCE ANY OUTPUT!
  config.keyUp = kUp;
  config.keyDown = kDown;
  config.keyLeft = kLeft;
  config.keyRight = kRight;
  config.keyInfo = kGreen;
  config.keyPageUp = kRed;
  config.keyPageDown = kYellow;
  config.keyScreenShot = kBlue;
  config.keyBack = kBack;
  config.keyOk = kOk;
}

cPluginFepg::~cPluginFepg()
{
  // Clean up after yourself!
}

const char *cPluginFepg::CommandLineHelp(void)
{
  // Return a string that describes all known command line options.
  return NULL;
}

bool cPluginFepg::ProcessArgs(int argc, char *argv[])
{
  // Implement command line argument processing here if applicable.
  return true;
}

bool cPluginFepg::Start(void)
{
  // Start any background activities the plugin shall perform.
  return true;
}

void cPluginFepg::Housekeeping(void)
{
  // Perform any cleanup or other regular tasks.
}

cOsdObject *cPluginFepg::MainMenuAction(void)
{
  // Perform the action when selected from the main VDR menu.
  return new cFEpg;
}

cMenuSetupPage *cPluginFepg::SetupMenu(void)
{
  // Return a setup menu in case the plugin supports one.
  return new cMenuSetupFEpg;
}

bool cPluginFepg::SetupParse(const char *Name, const char *Value)
{
  // Parse your own setup parameters and store their values.
  if      (!strcasecmp(Name, "keyUp"))         config.keyUp = ekeys[atoi(Value)];
  else if (!strcasecmp(Name, "keyDown"))       config.keyDown = ekeys[atoi(Value)];
  else if (!strcasecmp(Name, "keyLeft"))       config.keyLeft = ekeys[atoi(Value)];
  else if (!strcasecmp(Name, "keyRight"))      config.keyRight = ekeys[atoi(Value)];
  else if (!strcasecmp(Name, "keyInfo"))       config.keyInfo = ekeys[atoi(Value)];
  else if (!strcasecmp(Name, "keyPageUp"))     config.keyPageUp = ekeys[atoi(Value)];
  else if (!strcasecmp(Name, "keyPageDown"))   config.keyPageDown = ekeys[atoi(Value)];
  else if (!strcasecmp(Name, "keyScreenShot")) config.keyScreenShot = ekeys[atoi(Value)];
  else if (!strcasecmp(Name, "keyBack"))       config.keyBack = ekeys[atoi(Value)];
  else if (!strcasecmp(Name, "keyOk"))         config.keyOk = ekeys[atoi(Value)];
  
  else
     return false;
     
  return true;
}

sFEpgConfig config;

VDRPLUGINCREATOR(cPluginFepg); // Don't touch this!
